/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <sys/types.h>
#include <sys/wait.h>

#include <stdlib.h>
#include <dlfcn.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <string>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/cursorfont.h>

#include "nsCOMPtr.h"

// for plugins
#ifndef XP_UNIX
#include "nsString.h"
#include "nsIDOMNavigator.h"
#include "nsIDOMPluginArray.h"
#include "nsIDOMPlugin.h"
#include <plugin/nsIPluginHost.h>
#include "nsIDOMMimeType.h"
#include "nsIObserverService.h"

static NS_DEFINE_CID(kPluginManagerCID, NS_PLUGINMANAGER_CID);
#endif

#define DWORD uint32_t
#define BOOL bool
#define FALSE false
#define TRUE true

#include "controller.h"
#include <fstream>
#include "debug.h"

#include <log4cpp/PropertyConfigurator.hh>
#include <log4cpp/BasicConfigurator.hh>

const char* const LOGGER_CONFIG = "/usr/share/spice/logger.ini";
char gs_TrustStoreFile[32] = {0};
std::string conf_path;

#include "plugin.h"
#include "nsScriptablePeer.h"
#include "nsIServiceManager.h"
#include "nsISupportsUtils.h" // some usefule macros are defined here

#define MIME_TYPES_HANDLED  "application/x-spice"
#define PLUGIN_NAME         "Spice Firefox Plugin"
#define MIME_TYPES_DESCRIPTION  MIME_TYPES_HANDLED":qsc:"PLUGIN_NAME
#define PLUGIN_DESCRIPTION  PLUGIN_NAME " Spice Client wrapper for firefox"

char* NPP_GetMIMEDescription(void)
{
    return(const_cast<char*>(MIME_TYPES_DESCRIPTION));
}

//////////////////////////////////////
//
// general initialization and shutdown
//
NPError NS_PluginInitialize()
{
  return NPERR_NO_ERROR;
}

void NS_PluginShutdown()
{
}

// get values per plugin
NPError NS_PluginGetValue(NPPVariable aVariable, void *aValue)
{
  NPError err = NPERR_NO_ERROR;
  switch (aVariable) {
    case NPPVpluginNameString:
      *((char **)aValue) = const_cast<char*>(PLUGIN_NAME);
      break;
    case NPPVpluginDescriptionString:
      *((char **)aValue) = const_cast<char*>(PLUGIN_DESCRIPTION);
      break;
    default:
      err = NPERR_INVALID_PARAM;
      break;
  }
  return err;
}


/////////////////////////////////////////////////////////////
//
// construction and destruction of our plugin instance object
//
nsPluginInstanceBase * NS_NewPluginInstance(nsPluginCreateData * aCreateDataStruct)
{
    if(!aCreateDataStruct)
        return NULL;

    nsPluginInstance * plugin = new nsPluginInstance(aCreateDataStruct->instance);

    // now is the time to tell Mozilla that we are windowless
    NPN_SetValue(aCreateDataStruct->instance, NPPVpluginWindowBool, NULL);

    return plugin;
}

void NS_DestroyPluginInstance(nsPluginInstanceBase * aPlugin)
{
  if(aPlugin)
    delete (nsPluginInstance *)aPlugin;
}

////////////////////////////////////////
//
// nsPluginInstance class implementation
//
nsPluginInstance::nsPluginInstance(NPP aInstance) : nsPluginInstanceBase(),
  mInstance(aInstance),
  mInitialized(FALSE),
  mScriptablePeer(NULL)
{
    strncpy(mHomeDir, getenv("HOME"), sizeof(mHomeDir));
    strncat(mHomeDir, "/.spice", sizeof(mHomeDir));
    mkdir(mHomeDir, S_IRWXU);

    mString[0] = '\0';
    LOGGER_SECTION("spice.plugin");

    conf_path = LOGGER_CONFIG;

    std::ifstream log_init(conf_path.c_str());
    if (log_init.good()) {
        log_init.close();
        log4cpp::PropertyConfigurator::configure(conf_path.c_str());
    } else {
        log4cpp::BasicConfigurator::configure();
    }

    m_connected_status = -2;
}

nsPluginInstance::~nsPluginInstance()
{
    // mScriptablePeer may be also held by the browser 
    // so releasing it here does not guarantee that it is over
    // we should take precaution in case it will be called later
    // and zero its mPlugin member
    if (mScriptablePeer)
      NPN_ReleaseObject(mScriptablePeer);
}

// static void
// xt_event_handler(Widget xtwidget, nsPluginInstance *plugin, XEvent *xevent, Boolean *b)
// {
//   switch (xevent->type) {
//     case Expose:
//       // get rid of all other exposure events
//       if (plugin) {
//         //while(XCheckTypedWindowEvent(plugin->Display(), plugin->Window(), Expose, xevent));
//         plugin->draw();
//       }
//       default:
//         break;
//   }
// }

// void nsPluginInstance::draw()
// {
//   unsigned int h = mHeight/2;
//   unsigned int w = 3 * mWidth/4;
//   int x = (mWidth - w)/2; // center
//   int y = h/2;
//   XClearArea(mDisplay, mWindow, 0, 0, 0, 0, False);
//   if (x >= 0 && y >= 0) {
//     GC gc = XCreateGC(mDisplay, mWindow, 0, NULL);
//     if (!gc)
//       return;
//     XDrawRectangle(mDisplay, mWindow, gc, x, y, w, h);
//     if (mString[0]) {
//   int l = strlen(mString);
//   int fmba = mFontInfo->max_bounds.ascent;
//   int fmbd = mFontInfo->max_bounds.descent;
//   int fh = fmba + fmbd;
//   y += fh;
//   x += 32;
//       XDrawString(mDisplay, mWindow, gc, x, y, mString, l);
//     }
//     XFreeGC(mDisplay, gc);
//   }
// }

NPBool nsPluginInstance::init(NPWindow* aWindow)
{
    mInitialized = TRUE;

    memset (m_HostIP,0,256);
    memset (m_Port,0,256);
    memset (m_Password,0,256);
    memset (m_SecurePort,0,256);
    memset (m_CipherSuite,0,256);
    memset (m_SSLChannels,0,256);
    memset (m_TrustStore, 0 ,sizeof(m_TrustStore));
    memset (m_HostSubject, 0 ,sizeof(m_HostSubject));
    memset (m_Title,0,256);
    memset (m_DynamicMenu,0,256);
    memset (m_NumberOfMonitors,0,256);
    memset (m_GuestHostName,0,256);

    m_FullScreen = PR_FALSE;
    m_AdminConsole = PR_FALSE;
    m_NoTaskMgrExecution = PR_FALSE;
    m_SendCtrlAltdelete = PR_FALSE;
    m_UsbPort = 0;
    m_UsbAutoShare = PR_FALSE;

    return mInitialized;
}

NPError nsPluginInstance::SetWindow(NPWindow* aWindow)
{
    // keep window parameters
    mWindow = aWindow;
    return NPERR_NO_ERROR;
}


void nsPluginInstance::shut()
{
  mInitialized = FALSE;
}

NPBool nsPluginInstance::isInitialized()
{
  return mInitialized;
}

/* attribute string strHost; */
void nsPluginInstance::GetHostIP(char * *aHostIP) {
    strncpy (*aHostIP, m_HostIP, sizeof (*aHostIP));
}
void nsPluginInstance::SetHostIP(const char * aHostIP){
    strncpy(m_HostIP,aHostIP, sizeof (m_HostIP));
}

/* attribute string strPort; */
void nsPluginInstance::GetPort(char * *aPort) {
    strncpy (*aPort, m_Port, sizeof (*aPort));
}
void nsPluginInstance::SetPort(const char * aPort) {
    strncpy(m_Port,aPort, sizeof (m_Port));
}

/* attribute ing SecurePort; */
void nsPluginInstance::GetSecurePort(char * *aSecurePort) {
    strncpy (*aSecurePort, m_SecurePort, sizeof (*aSecurePort));
}
void nsPluginInstance::SetSecurePort(const char * aSecurePort) {
    strncpy(m_SecurePort,aSecurePort, sizeof (m_SecurePort));
}


/* attribute string strPassword; */
void nsPluginInstance::GetPassword(char * *aPassword) {
    strncpy (*aPassword, m_Password, sizeof (*aPassword));
}
void nsPluginInstance::SetPassword(const char * aPassword) {
    strncpy(m_Password,aPassword, sizeof (m_Password));
}

/* attribute ing Port; */
void nsPluginInstance::GetCipherSuite(char * *aCipherSuite) {
    strncpy (*aCipherSuite, m_CipherSuite, sizeof (*aCipherSuite));
}
void nsPluginInstance::SetCipherSuite(const char * aCipherSuite) {
    strncpy(m_CipherSuite,aCipherSuite, sizeof (m_CipherSuite));
}

/* attribute string strPort; */
void nsPluginInstance::GetSSLChannels(char * *aSSLChannels) {
    strncpy (*aSSLChannels, m_SSLChannels, sizeof (*aSSLChannels));
}
void nsPluginInstance::SetSSLChannels(const char * aSSLChannels) {
    strncpy(m_SSLChannels,aSSLChannels, sizeof (m_SSLChannels));
}

//* attribute string TrustStore; */
void nsPluginInstance::GetTrustStore(char * *aTrustStore) {
    strncpy (*aTrustStore, m_TrustStore, sizeof (*aTrustStore));
}
void nsPluginInstance::SetTrustStore(const char * aTrustStore) {
    strncpy(m_TrustStore, aTrustStore, sizeof (m_TrustStore));
}

//* attribute string HostSubject; */
void nsPluginInstance::GetHostSubject(char * *aHostSubject) {
    strncpy (*aHostSubject, m_HostSubject, sizeof (*aHostSubject));
}
void nsPluginInstance::SetHostSubject(const char * aHostSubject) {
    strncpy(m_HostSubject, aHostSubject, sizeof (m_HostSubject));
}

/* attribute string strFullScreen; */
void nsPluginInstance::GetFullScreen(PRBool *aFullScreen) {
    *aFullScreen = m_FullScreen;
}
void nsPluginInstance::SetFullScreen(PRBool aFullScreen) {
    m_FullScreen = aFullScreen;
}

/* attribute string strTitle; */
void nsPluginInstance::GetTitle(char * *aTitle) {
    strncpy (*aTitle, m_Title, sizeof (*aTitle));
}
void nsPluginInstance::SetTitle(const char * aTitle) {
    strncpy(m_Title,aTitle, sizeof (m_Title));
}

/* attribute string strDynamicMenu; */
void nsPluginInstance::GetDynamicMenu(char * *aDynamicMenu) {
    strncpy (*aDynamicMenu, m_DynamicMenu, sizeof (*aDynamicMenu));
}
void nsPluginInstance::SetDynamicMenu(const char * aDynamicMenu) {
    strncpy(m_DynamicMenu,aDynamicMenu, sizeof (m_DynamicMenu));
}

/* attribute string strNumberOfMonitors; */
void nsPluginInstance::GetNumberOfMonitors(char * *aNumberOfMonitors) {
    strncpy (*aNumberOfMonitors, m_NumberOfMonitors, sizeof (*aNumberOfMonitors));
}
void nsPluginInstance::SetNumberOfMonitors(const char * aNumberOfMonitors) {
    strncpy(m_NumberOfMonitors,aNumberOfMonitors, sizeof (m_NumberOfMonitors));
}

/* attribute string strAdminConsole; */
void nsPluginInstance::GetAdminConsole(PRBool *aAdminConsole) {
    *aAdminConsole = m_AdminConsole;
}
void nsPluginInstance::SetAdminConsole(PRBool aAdminConsole) {
    m_AdminConsole = aAdminConsole;
}

/* attribute string strGuestHostName; */
void nsPluginInstance::GetGuestHostName(char * *aGuestHostName) {
    strncpy (*aGuestHostName, m_GuestHostName, sizeof (*aGuestHostName));
}
void nsPluginInstance::SetGuestHostName(const char * aGuestHostName) {
    strncpy(m_GuestHostName,aGuestHostName, sizeof (m_GuestHostName));
}

/* attribute string strHotKeys; */
void nsPluginInstance::GetHotKeys(char * *aHotKeys) {
    strncpy (*aHotKeys, m_HotKeys, sizeof (*aHotKeys));
}
void nsPluginInstance::SetHotKeys(const char * aHotKeys) {
    strncpy(m_HotKeys,aHotKeys, sizeof (m_HotKeys));
}

/* attribute string strNoTaskMgrExecution; */
void nsPluginInstance::GetNoTaskMgrExecution(PRBool *aNoTaskMgrExecution) {
    *aNoTaskMgrExecution = m_NoTaskMgrExecution;
}
void nsPluginInstance::SetNoTaskMgrExecution(PRBool aNoTaskMgrExecution) {
    m_NoTaskMgrExecution = aNoTaskMgrExecution;
;
}

/* attribute string strSendCtrlAltdelete; */
void nsPluginInstance::GetSendCtrlAltdelete(PRBool *aSendCtrlAltdelete) {
    *aSendCtrlAltdelete = m_SendCtrlAltdelete;
}
void nsPluginInstance::SetSendCtrlAltdelete(PRBool aSendCtrlAltdelete) {
    m_SendCtrlAltdelete = aSendCtrlAltdelete;
}

/* attribute string UsbListenPort; */
void nsPluginInstance::GetUsbListenPort(unsigned short *aUsbPort) {
    if (aUsbPort != NULL) {
        *aUsbPort = m_UsbPort;
    }
}

void nsPluginInstance::SetUsbListenPort(unsigned short aUsbPort) {
    m_UsbPort = aUsbPort;
}

/* attribute boolean UsbAutoShare; */
void nsPluginInstance::GetUsbAutoShare(PRBool *aUsbAutoShare) {
    if (aUsbAutoShare != NULL) {
        *aUsbAutoShare = m_UsbAutoShare;
    }
}

void nsPluginInstance::SetUsbAutoShare(PRBool aUsbAutoShare) {
    m_UsbAutoShare = aUsbAutoShare;
}

void *WaitThread (void *opaque) {
    int nRC;
    nsresult rv;
    nsPluginInstance *plugin = (nsPluginInstance *)opaque;

    waitpid(plugin->m_child_pid, &nRC, 0);
    LOG_DEBUG("spicec exit code = " << nRC);
    plugin->m_connected_status = plugin->m_external_controller->TranslateRC(nRC);
    unlink (gs_TrustStoreFile);
    memset(gs_TrustStoreFile, 0 ,sizeof(gs_TrustStoreFile));
    return NULL;
}

void nsPluginInstance::write_to_pipe(const void * data, uint32_t size)
{
    uint32_t written;
    m_external_controller->Write(data, size, &written);
}

void nsPluginInstance::send_init()
{
    ControllerInit msg = {{CONTROLLER_MAGIC, CONTROLLER_VERSION, sizeof(msg)}, 0,
                          CONTROLLER_FLAG_EXCLUSIVE};
    write_to_pipe(&msg, sizeof(msg));
}

void nsPluginInstance::send_msg(uint32_t id)
{
    ControllerMsg msg = {id, sizeof(msg)};
    write_to_pipe(&msg, sizeof(msg));
}

void nsPluginInstance::send_value(uint32_t id, uint32_t value)
{
    if (!value) {
        return;
    }
    ControllerValue msg = {{id, sizeof(msg)}, value};
    write_to_pipe(&msg, sizeof(msg));
}

void nsPluginInstance::send_str(uint32_t id, char* str)
{
    if (!strlen(str)) {
        return;
    }
    size_t size = sizeof(ControllerData) + strlen(str) + 1;
    ControllerData* msg = (ControllerData*)malloc(size);
    msg->base.id = id;
    msg->base.size = size;
    strcpy((char*)msg->data, str);
    write_to_pipe(msg, size);
    free(msg);
}

void nsPluginInstance::send_wstr(uint32_t id, wchar_t* str)
{
    size_t len = wcslen(str);
    if (!len) {
        return;
    }
    size_t size = sizeof(ControllerData) + (len + 1) * sizeof(wchar_t);
    ControllerData* msg = (ControllerData*)malloc(size);
    msg->base.id = id;
    msg->base.size = size;
    wcscpy((wchar_t*)msg->data, str);
    write_to_pipe(msg, size);
    free(msg);
}



void nsPluginInstance::Connect() {
    nsresult rv;

    MessageData mdData;
    pid_t pid = getpid();
    timespec now;

    mdData.nReadPos = 0;
    mdData.nCurrentReadLength = 4;

    // Place socket file to ~/.spice/spice-XXXXXX
    strncpy(mdData.name, mHomeDir, sizeof(mHomeDir));
    strcat(mdData.name, "/spice-XXXXXX");
    mktemp(mdData.name);
    if (setenv("SPICE_XPI_SOCKET", mdData.name, 1)) {
        LOG_ERROR ("could not set SPICE_XPI_SOCKET env variable");
        return;
    }

    m_child_pid = fork();
    LOG_DEBUG(" m_child_pid = " << m_child_pid);
    if (m_child_pid == 0) {
        execl (
            "/usr/libexec/spicec",
            "/usr/libexec/spicec",
            "--controller",
            NULL
        );
        // failed to connect?
        exit(0);
    } else {
        m_external_controller = new SpiceController(mdData);

        if (m_external_controller->Connect(10) != 0) {
            LOG_ERROR ("could not connect to spice client controller");
            return;
        }

        int nSpicecRC;
        pthread_t wait_thread_id;

        // create TruststoreFilename
        FILE *fp;
        int fd = -1;
        strncpy (
            gs_TrustStoreFile,
            "/tmp/truststore.pem-XXXXXX",
            sizeof(gs_TrustStoreFile)
        );
        fd = mkstemp(gs_TrustStoreFile);
        if (fd != -1) {
            fp = fdopen (fd,"w+");
            if (fp != NULL) {
                fputs(m_TrustStore, fp);
                fflush(fp);
                fsync(fd);
                fclose(fp);
            } else {
                LOG_ERROR ("could not open truststore temp file");
                close(fd);
                unlink(gs_TrustStoreFile);
                memset(gs_TrustStoreFile, 0 ,sizeof(gs_TrustStoreFile));
                return;
            }
        } else {
            LOG_ERROR ("could not create truststore temp file. errno=" << errno);
            return;
        }

        send_init();
        send_str(CONTROLLER_HOST, m_HostIP);
        send_value(CONTROLLER_PORT, atoi(m_Port));
        send_value(CONTROLLER_SPORT, atoi(m_SecurePort));
        send_value(CONTROLLER_FULL_SCREEN,
                   (m_FullScreen == PR_TRUE ? CONTROLLER_SET_FULL_SCREEN : 0) |
                   (m_AdminConsole == PR_FALSE ? CONTROLLER_AUTO_DISPLAY_RES : 0));
        send_str(CONTROLLER_PASSWORD, m_Password);
        send_str(CONTROLLER_TLS_CIPHERS, m_CipherSuite);
        send_str(CONTROLLER_SET_TITLE, m_Title);

        /*
         * HACK -- remove leading s from m_SSLChannels, e.g. "main" not "smain"
         * RHEL5 uses 'smain' and 'sinpusts
         * RHEL6 uses 'main'  and 'inputs'
         */
        if (1) {
            int i;
            for (i=0; i<sizeof(m_SSLChannels); i++) {
                char *p = &m_SSLChannels[i];
                if (!*p)
                    break;
                if ((strncmp("smain",   p, 5) == 0) ||
                    (strncmp("sinputs", p, 7) == 0)) {
                    while ((*p = *(p+1)) != '\0') /* copy is done in condition */
                    p++;
                 }
            }
        } /* HACK */

        send_str(CONTROLLER_SECURE_CHANNELS, m_SSLChannels);
        send_str(CONTROLLER_CA_FILE, gs_TrustStoreFile);
        send_str(CONTROLLER_HOST_SUBJECT, m_HostSubject);
        send_str(CONTROLLER_HOTKEYS, m_HotKeys);
        send_msg(CONTROLLER_CONNECT);
        send_msg(CONTROLLER_SHOW);

        // Set connected status
        m_connected_status = -1;

        // Setup wait thread
        pthread_create (&wait_thread_id, NULL, WaitThread, (void*)this);

        ExecuteUsbCtrl();
    }
}

void nsPluginInstance::show() {
    LOG_DEBUG("sending show message");
    send_msg(CONTROLLER_SHOW);
}

void nsPluginInstance::disconnect() {
    if (m_child_pid != 0) {
        kill (m_child_pid, SIGTERM);
    }
}

void nsPluginInstance::ConnectedStatus(PRInt32 *_retval) {
    *_retval = m_connected_status;
}

void nsPluginInstance::SetLanguageStrings(const char * aSection, const char * aLanguage) {
    if ((aSection != NULL) && (aLanguage != NULL)) {
        if ((strlen(aSection) > 0) && (strlen(aLanguage) > 0)) {
            m_Language[aSection] = aLanguage;
        }   
    }
}

void nsPluginInstance::SetUsbFilter(const char * aUsbFilter) {
    if (aUsbFilter != NULL) {
        m_UsbFilter = aUsbFilter;
    }
}

void nsPluginInstance::ExecuteUsbCtrl()
{
    if ((m_child_pid == 0) || (m_UsbPort == 0)) {
        return;
    }

    LOG_INFO("Executing usbrdrctrl application.");

    m_UsbCtrlPid = fork();

    if (m_UsbCtrlPid == 0) {
    
        char logname[64];
        snprintf(logname, sizeof(logname), "/tmp/usbrdrctrl-%d.log", getpid());
        
        int logfd = open(logname, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
        if (logfd > 0) {
            dup2(logfd, 1);
        }
        
        char conn_name[64];
        snprintf(conn_name, sizeof(conn_name), "/tmp/SpiceForeignMenu-%d.uds", m_child_pid);
        
        LOG_DEBUG("Connection name: " << conn_name);

        char port[6];
        snprintf(port, sizeof(port), "%d", m_UsbPort);

        sleep(2);
        
        // TBD: Find executable path.
        execl("/usr/bin/usbrdrctrl",
            "/usr/bin/usbrdrctrl",
            m_GuestHostName,
            port,
            "-c",
            conn_name,
            !m_Language["USB"].empty() ? "-l" : "",
            !m_Language["USB"].empty() ? m_Language["USB"].c_str() : "",
            !m_UsbFilter.empty() ? "-f" : "",
            !m_UsbFilter.empty() ? m_UsbFilter.c_str() : "",
            (m_UsbAutoShare == PR_TRUE) ? "-a" : "",
            NULL);

        LOG_ERROR("Failed to execute usbrdrctrl: " << errno << ": " << strerror(errno));
        exit(0);
            
    } else if (m_UsbCtrlPid > 0) {
        pthread_t tid;
        pthread_create (&tid, NULL, UsbCtrlWatchDog, this);
    }
}

void* nsPluginInstance::UsbCtrlWatchDog(void *opaque)
{
    nsPluginInstance *pThis = reinterpret_cast<nsPluginInstance*>(opaque);
    
    if (pThis != NULL) {
        int exitcode;
        
        waitpid(pThis->m_UsbCtrlPid, &exitcode, 0);
        LOG_DEBUG("usbctrl was terminated with exit code " << exitcode);
        
        pThis->m_UsbCtrlPid = 0;
        
        if (exitcode != 0) {
            pThis->ExecuteUsbCtrl();
        }
    }
    
    return NULL;
}

// ==============================
// ! Scriptability related code !
// ==============================
//
// here the plugin is asked by Mozilla to tell if it is scriptable
// we should return a valid interface id and a pointer to 
// nsScriptablePeer interface which we should have implemented
// and which should be defined in the corressponding *.xpt file
// in the bin/components folder
NPError nsPluginInstance::GetValue(NPPVariable aVariable, void *aValue)
{
  NPError rv = NPERR_NO_ERROR;

  // Here we indicate that the plugin is scriptable. See this page for details:
  // https://developer.mozilla.org/en/Gecko_Plugin_API_Reference/Scripting_plugins
  if (aVariable == NPPVpluginScriptableNPObject) {
      // addref happens in getter, so we don't addref here
    *(NPObject **)aValue = getScriptablePeer();
  }
  return rv;
}

// ==============================
// ! Scriptability related code !
// ==============================
//
// this method will return the scriptable object (and create it if necessary)
NPObject* nsPluginInstance::getScriptablePeer()
{
  NPIdentifiersInit();

  if (!mScriptablePeer) {
    mScriptablePeer = NPN_CreateObject(mInstance, GET_NPOBJECT_CLASS(ScriptablePluginObject)); 
  }
  
  if (mScriptablePeer) {
    NPN_RetainObject(mScriptablePeer);
  } 
    
  return mScriptablePeer;
}
